#!/usr/bin/env python

__revision__ = '$Id: zcfg_addon.py 10 2005-03-07 22:05:31Z nl $'

versionInfo = (0, 2)
__version__ = '.'.join([str(num) for num in versionInfo])

__docformat__ = 'restructuredtext en'

import sys

sys.modules[__name__].__doc__ = """
ZConfig extension that allows one to modify and save the configuration.

The `ZConfig`__ library parses a config file, validating it against a schema,
and returns a hierarchy of objects with matching keys and values.

.. __: http://www.zope.org/Members/fdrake/zconfig/

There are a number of use cases where the config values have to be modified,
and the resulting config has to be saved to file. In addition, it is
sometimes convenient to convert the hierarchy of objects representing the
config, to and from a hierarchy of nested dictionaries.

This extension implements such functionality by adding, at runtime, five
methods, one function and one class to the imported ZConfig modules and
classes, avoiding changes to the ZConfig source code.

It has been tested using ZConfig v.2.2 and Python v.2.3.

See the *test* function source code for examples of usage of the methods,
function and class.

:author: Nicola Larosa
:contact: <nico@tekNico.net>
:license: MIT license; see LICENSE file
:version: %(__version__)s

:Variables:
  - `__revision__`: summary of metadata from the Subversion repository
  - `versionInfo`: tuple with current major and minor version numbers
  - `__version__`: string computed from the *versionInfo* value
  - `TRUE_STR`: True bool value in generated text, may be ``yes``, ``on``,
    ``true``
  - `FALSE_STR`: False bool value in generated text, may be ``no``, ``off``,
    ``false``
""" % vars()

import copy, difflib
from pprint import pprint
from StringIO import StringIO

import ZConfig
if not (ZConfig.version_info >= (2, 2)):
    raise ImportError('zcfg_addon needs ZConfig version 2.2 or higher')

# the addons to ZConfig are placed before the test function

TRUE_STR = 'yes'
FALSE_STR = 'no'

def _getChildInfo(self,
        name):
    """
    Find the *childInfo* object for *name* and return it.
    
    :Parameters:
      - `name`: name of child whose *childInfo* object is to be found
    
    :return: the *childInfo* object associated with *name*
    :rtype: *childInfo*
    
    :Exceptions:
      - `AssertionError`: raised if no child of self named *name* can be found
    
    Method added to the *ZConfig.matcher.SectionValue* class.
    """
    childInfo = None
    for _, curCI in self._matcher.type:
        if curCI.attribute == name:
            childInfo = curCI
            break
    assert childInfo
    return childInfo

def addSectionToMultisection(self,
        typeName,
        valDict):
    """
    Add a flat section (no subsections) to a *Multisection* object.
    
    :Parameters:
      - `typeName`: name of the type of the *Multisection*
      - `valDict`: dict of values to be used in the section to be added
    
    :return: None
    
    :Exceptions:
      - `AssertionError`: raised if *typeName* is not an attribute of self,
        or if its *childInfo* is not a *Multisection*
    
    Method added to the *ZConfig.matcher.SectionValue* class.
    """
    assert typeName in self._attributes
    childInfo = self._getChildInfo(typeName)
    assert childInfo.issection() and childInfo.ismulti()
    subType = self._matcher.type.gettype(typeName)
    subMatcher = self._matcher.createChildMatcher(subType, '')
    for key, val in valDict.items():
        # fake not existent (row, col, url) tuple
        subMatcher.addValue(key, val, (0, 0, ''))
    sectVal = subMatcher.finish()
    sectVal = sectVal._matcher.type.datatype(sectVal)
    self._matcher.addSection(typeName, '', sectVal)
    return

def delSectionFromMultisection(self,
        typeName,
        idx):
    """
    Delete a section from a Multisection object.
    
    :Parameters:
      - `typeName`: name of the type of the *Multisection*
      - `idx`: index of the section to be deleted
    
    :return: None
    
    :Exceptions:
      - `AssertionError`: raised if *typeName* is not an attribute of self,
        or if its *childInfo* is not a *Multisection*
    
    Method added to the *ZConfig.matcher.SectionValue* class.
    """
    assert typeName in self._attributes
    childInfo = self._getChildInfo(typeName)
    assert childInfo.issection() and childInfo.ismulti()
    multi = getattr(self, typeName)
    del multi[idx]
    return

def addValueToMultikey(self,
        name,
        val):
    """
    Add one value to a Multikey object.
    
    :Parameters:
      - `name`: name of the *Multikey*
      - `val`: value to be added
    
    :return: None
    
    :Exceptions:
      - `AssertionError`: raised if *name* is not an attribute of self,
        or if its *childInfo* is not a *Multikey*
    
    Method added to the *ZConfig.matcher.SectionValue* class.
    """
    assert name in self._attributes
    childInfo = self._getChildInfo(name)
    assert not childInfo.issection() and childInfo.ismulti()
    # fake not existent (row, col, url) tuple
    self._matcher.addValue(name, val, (0, 0, ''))
    valueInfo = self._matcher._values[name][-1]
    self._matcher._values[name][-1] = valueInfo.convert(childInfo.datatype)
    return

def delValueFromMultikey(self,
        name,
        idx):
    """
    Delete a value from a Multikey object.
    
    :Parameters:
      - `name`: name of the *Multikey*
      - `idx`: index of value to be deleted
    
    :return: None
    
    :Exceptions:
      - `AssertionError`: raised if *name* is not an attribute of self,
        or if its *childInfo* is not a *Multikey*
    
    Method added to the *ZConfig.matcher.SectionValue* class.
    """
    assert name in self._attributes
    childInfo = self._getChildInfo(name)
    assert not childInfo.issection() and childInfo.ismulti()
    multi = getattr(self, name)
    del multi[idx]
    return

def modSingleValue(self,
        name,
        val):
    """
    Modify a simple value.
    
    :Parameters:
      - `name`: name of the *Key*
      - `val`: value to be modified
    
    :return: None
    
    :Exceptions:
      - `AssertionError`: raised if *name* is not an attribute of self,
        or if its *childInfo* is not a simple *Key*
    
    Method added to the *ZConfig.matcher.SectionValue* class.
    """
    assert name in self._attributes
    childInfo = self._getChildInfo(name)
    assert not childInfo.issection() and not childInfo.ismulti()
    # fake not existent (row, col, url) tuple
    valueInfo = ZConfig.info.ValueInfo(val, (0, 0, ''))
    value = valueInfo.convert(childInfo.datatype)
    self._matcher._values[name] = self.__dict__[name] = value
    return

def toDict(self,
        sectName = None):
    """
    Convert a *SectionValue* object to a dictionary.
    
    :Parameters:
      - `sectName`: optional, name of the section if different from the type
        name
    
    :return: a tree of dictionaries containing the section values
    :rtype: dict
    
    Recursive method added to the *ZConfig.matcher.SectionValue* class.
    """
    retDict = {}
    if sectName is not None:
        retDict['__sect_name__'] = sectName
    for key, subInfo in self.getSectionDefinition():
        attr = subInfo.attribute
        value = getattr(self, attr)
        if subInfo.issection():
            # value is a section
            if subInfo.ismulti():
                retDict[attr] = [subSect.toDict(subSect.getSectionName())
                    for subSect in value if subSect] or None
            else:
                retDict[attr] = value and value.toDict() or None
        else:
            if subInfo.ismulti():
                retDict[attr] = [val for val in value] or None
            else:
                retDict[attr] = value
    return retDict

def fromDict(
        schema,
        cfgDict):
    """
    Build a *SectionValue* config object from a conforming dictionary.
    
    :Parameters:
      - `schema`: a *SchemaType* object
      - `cfgDict`: a dictionary tree, as returned by the toDict method
    
    :return: a *SectionValue* object with the same values as cfgDict
    :rtype: *SectionValue*
    
    Factory function added to the ZConfig.loader module.
    """
    section = ZConfig.matcher.SchemaMatcher(schema)
    _addSectFromDict(schema, section, cfgDict)
    return section.finish()

def _addSectFromDict(
        schema,
        parent,
        cfgDict):
    """
    Create a *SectionValue* object getting values from a dictionary, and
    append it to a *SectionValue* parent object.

    :Parameters:
      - `schema`: a *SchemaType* object
      - `parent`: a *SectionValue* object to append the built section to
      - `cfgDict`: a dictionary tree, as returned by the toDict method
    
    :return: None
    
    Recursive helper function.
    """
    for key, subInfo in parent.type:
        attr = subInfo.attribute
        if cfgDict.has_key(attr) and (cfgDict[attr] is not None):
            if subInfo.issection():
                sectTypeName = subInfo.sectiontype.name
                sectType = parent.type.gettype(sectTypeName)
                if subInfo.ismulti():
                    for value in cfgDict[attr]:
                        if value is not None:
                            subSect = parent.createChildMatcher(
                                sectType,
##                                (sectTypeName != attr) and attr or '')
                                value.get('__sect_name__', ''))
                            _addSectFromDict(schema, subSect, value)
                            sectValue = subSect.finish()
                            parent.addSection(sectTypeName, '', sectValue)
                else:
                    subSect = parent.createChildMatcher(sectType, attr)
                    _addSectFromDict(schema, subSect, cfgDict[attr])
                    sectValue = subSect.finish()
                    parent.addSection(sectTypeName, attr, sectValue)
            else:
                if subInfo.ismulti():
                    for value in cfgDict[attr]:
                        if value is not None:
                            # fake not existent (row, col, url) tuple
                            parent.addValue(attr, str(value),
                                (0, 0, ''))
                else:
                    # fake not existent (row, col, url) tuple
                    parent.addValue(attr, str(cfgDict[attr]), (0, 0, ''))
    return

class ZConfigGenerator:
    """
    ZConfig config file generator extension.
    
    Use the *createSectionLines* method to generate the config text lines
    for a *SectionValue* object.
    
    :Cvariables:
      - `indentStr`: blank string added for each indentation level
    
    Class added to the *ZConfig.cfgparser* module.
    """
    
    __metaclass__ = type
    indentStr = '  '
    
    def createSectionLines(self,
            sect,
            indentLevel = 0):
        """
        Generate config section lines for a *SectionValue* object.
        
        :Parameters:
          - `sect`: a *SectionValue* object
          - `indentLevel`: the indentation level of the generated lines
        
        :return: a list of strings with the generated config lines
        :rtype: list
        """
        cfgLines = []
        for key, subInfo in sect.getSectionDefinition():
            value = getattr(sect, subInfo.attribute)
            if subInfo.issection():
                # value is a section
                typeName = subInfo.sectiontype.name
                if subInfo.ismulti():
                    # value is a sequence
                    for subSect in value:
                        if subSect:
                            subLines = self._appendSubsectLines(
                                typeName,
                                subSect.getSectionName() or '',
                                subSect,
                                indentLevel)
                            cfgLines.extend(subLines)
                        else:
                            # do not write anything for empty sessions
                            pass
                else:
                    if value:
                        sectName = subInfo.attribute
                        subLines = self._appendSubsectLines(
                            typeName,
                            (typeName != sectName) and sectName or '',
                            value,
                            indentLevel)
                        cfgLines.extend(subLines)
                    else:
                        # do not write anything for empty sessions
                        pass
            else:
                indentStr = self.indentStr * indentLevel
                if subInfo.ismulti():
                    for val in value:
                        if val is not None:
                            if isinstance(val, bool):
                                val = val and TRUE_STR or FALSE_STR
                            cfgLines.append('%s%s %s' % (indentStr, key, val))
                else:
                    if value is not None:
                        if isinstance(value, bool):
                            value = value and TRUE_STR or FALSE_STR
                        cfgLines.append('%s%s %s' % (indentStr, key, value))
        return cfgLines

    def _appendSubsectLines(self,
            typeName,
            sectName,
            subSect,
            indentLevel):
        """
        Append config lines for a subsection.
        
        :Parameters:
          - `typeName`: name of the section type
          - `sectName`: optional, name of the section if different from the
            type name
          - `subSect`: a *SectionValue* object for the subsection
          - `indentLevel`: the indentation level of the generated lines
        
        :return: a list of strings with the generated config lines
        :rtype: list
        
        Helper method.
        """
        cfgLines = []
        indentStr = self.indentStr * indentLevel
        # start of subsection
        line = '%s<%s' % (indentStr, typeName)
        # sectName may be empty
        if sectName:
            line = '%s %s>' % (line, sectName)
        else:
            line += '>'
        cfgLines.append(line)
        # recurse
        subLines = self.createSectionLines(
            subSect,
            indentLevel + 1)
        cfgLines.extend(subLines)
        # end of subsection
        cfgLines.append('%s</%s>' % (indentStr, typeName))
        return cfgLines

# ZConfig.matcher.SectionValue addons
from ZConfig.matcher import SectionValue
SectionValue._getChildInfo              = _getChildInfo
SectionValue.addSectionToMultisection   = addSectionToMultisection
SectionValue.delSectionFromMultisection = delSectionFromMultisection
SectionValue.addValueToMultikey         = addValueToMultikey
SectionValue.delValueFromMultikey       = delValueFromMultikey
SectionValue.modSingleValue             = modSingleValue
SectionValue.toDict                     = toDict

# ZConfig.loader addon
ZConfig.loader.fromDict = fromDict

# ZConfig.cfgparser addon
ZConfig.cfgparser.ZConfigGenerator = ZConfigGenerator

def test():
    """
    Main testing function, invoked when the script is launched.
    
    Contains usage examples of the added methods, class and function.
    """

    testSchema = (
"""<schema>

  <sectiontype name="third_level">
    <key name="tl_str" datatype="string" required="yes"/>
    <key name="tl_int" datatype="integer" default="3"/>
  </sectiontype>

  <sectiontype name="second_level">
    <key name="sl_str" datatype="string" required="yes"/>
    <key name="sl_int" datatype="integer" default="2"/>
    <multisection name="*" type="third_level" attribute="third_level"/>
  </sectiontype>

  <sectiontype name="first_level">
    <key name="fl_key" datatype="boolean" default="False"/>
    <section name="fl_sec" type="second_level" required="yes"/>
    <multisection name="*" type="second_level" attribute="fl_ms"/>
  </sectiontype>

  <sectiontype name="standalone">
    <key name="sa_str" datatype="string"/>
    <multikey name="sa_mk" datatype="string"/>
  </sectiontype>

<!-- section instances -->

  <section name="*" type="first_level" attribute="first_level"
    required="yes"/>

  <section name="*" type="standalone" attribute="standalone"/>

</schema>
""")
    testSchemaFile = StringIO(testSchema)

    origConfig = (
"""<first_level>
  fl_key no
  <second_level fl_sec>
    sl_str 2ndLevStr3
    sl_int 5
    <third_level>
      tl_str 3rdLevStr3
      tl_int 8
    </third_level>
    <third_level>
      tl_str 3rdLevStr11
      tl_int 31
    </third_level>
  </second_level>
  <second_level fl_ms>
    sl_str 2ndLevStr4
    sl_int 1
    <third_level>
      tl_str 3rdLevStr7
      tl_int 3
    </third_level>
    <third_level>
      tl_str 3rdLevStr15
      tl_int 24
    </third_level>
  </second_level>
</first_level>
<standalone>
  sa_str saStr
  sa_mk saMk1
  sa_mk saMk2
</standalone>
""")
    origConfigFile = StringIO(origConfig)
    origConfigList = origConfig.split('\n')

    modifConfig = (
"""<first_level>
  fl_key no
  <second_level fl_sec>
    sl_str 2ndLevStr3
    sl_int 11
    <third_level>
      tl_str 3rdLevStr11
      tl_int 31
    </third_level>
  </second_level>
  <second_level fl_ms>
    sl_str 2ndLevStr4
    sl_int 1
    <third_level>
      tl_str 3rdLevStr7
      tl_int 3
    </third_level>
    <third_level>
      tl_str 3rdLevStr15
      tl_int 24
    </third_level>
    <third_level>
      tl_str added
      tl_int 42
    </third_level>
  </second_level>
</first_level>
<standalone>
  sa_str saStr
  sa_mk saMk1
  sa_mk added
</standalone>
""")
    modifConfigList = modifConfig.split('\n')

    cfgGenerator = ZConfig.cfgparser.ZConfigGenerator()

    def configLines(
            cfg):
        """
        Generate a config section text.
        """
        cfgLines = cfgGenerator.createSectionLines(cfg)
        # add the final newline
        cfgLines.append('')
        return '\n'.join(cfgLines)

    # create config object
    schema = ZConfig.loadSchemaFile(testSchemaFile)
    origCfg, _ = ZConfig.loadConfigFile(schema, origConfigFile)
    # check that the same unmodified config text is generated
    uniDiff = difflib.unified_diff(
        origConfigList,
        configLines(origCfg).split('\n'))
    print 'Diff between original and text from the umodified object:'
    pprint(list(uniDiff))
    # check the config text after conversion to and from dict
    dictCfg = ZConfig.loader.fromDict(schema, origCfg.toDict())
    uniDiff = difflib.unified_diff(
        origConfigList,
        configLines(dictCfg).split('\n'))
    print 'Diff between original and object to and from dict:'
    pprint(list(uniDiff))
    # modify config
    modifCfg = copy.copy(origCfg)
    sectVal = {'tl_str': 'added', 'tl_int': 42}
    # add another third_level section after the others
    modifCfg.first_level.fl_ms[0].addSectionToMultisection(
        'third_level',
        sectVal)
    # delete the first third_level section
    modifCfg.first_level.fl_sec.delSectionFromMultisection(
        'third_level',
        0)
    # add a value to the sa_mk multikey after the others
    modifCfg.standalone.addValueToMultikey(
        'sa_mk',
        'added')
    # delete the second value from the sa_mk multikey
    modifCfg.standalone.delValueFromMultikey(
        'sa_mk',
        1)
    # modify the sl_int single value from 5 to 11
    modifCfg.first_level.fl_sec.modSingleValue(
        'sl_int',
        11)
    # check that the modified config text is as wanted
    uniDiff = difflib.unified_diff(
        modifConfigList,
        configLines(modifCfg).split('\n'))
    print 'Diff between wanted modified text and text from modified object:'
    pprint(list(uniDiff))

if __name__ == '__main__':
    test()

